import sys
import os
import numpy as np
from PySide6.QtWidgets import (QApplication, QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QListWidget, QSpinBox,QDoubleSpinBox, QLabel, QCheckBox, QFileDialog, QTabWidget, QSizePolicy, QGridLayout, QSpinBox, QMessageBox, QLineEdit, QFormLayout, QHBoxLayout, QStyleFactory,)
from PySide6.QtCore import Qt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from functools import partial
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter
from PySide6.QtGui import QPalette, QColor, QFont
import math

class MplCanvas(FigureCanvas):
    def __init__(self, width=5, height=4, dpi=300):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super().__init__(fig)

class PlottingApplication(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("SD Frequency and multi-plots")
        self.histogram_bins = 100  # Default number of bins
        self.scatter_marker_size = 3  # Default scatter marker size
        self.global_limits = {}
        self.initUI()

    def initUI(self):
        # Main widget and layout
        mainWidget = QWidget()
        self.setCentralWidget(mainWidget)
        mainLayout = QHBoxLayout(mainWidget)

        # Left panel for file selection and plot options
        self.leftPanel = QTabWidget()  # Assign to self to make it accessible class-wide
        fileSelectionTab = QWidget()
        self.plotOptionsTab = QWidget()
        self.leftPanel.addTab(fileSelectionTab, "Files")
        self.leftPanel.addTab(self.plotOptionsTab, "Options")
        
        # Right panel for displaying plots
        self.rightPanel = QTabWidget()

        mainLayout.addWidget(self.leftPanel, 25)
        mainLayout.addWidget(self.rightPanel, 75)

        self.setupFileSelectionTab(fileSelectionTab)
        self.setupPlotOptionsTab(self.plotOptionsTab)
    


    def calculateGridSize(self, numberOfPlots):
        # Calculate the square root of the number of plots to get an approximation
        # of the grid size
        gridSize = math.ceil(math.sqrt(numberOfPlots))
        
        # For a perfectly square layout, both the number of rows and columns
        # would be equal to gridSize. However, depending on your exact needs,
        # you might want to adjust this to have more rows than columns or vice versa.
        nrows = gridSize
        ncols = gridSize
        
        # Adjust ncols if there are too many empty spots
        if gridSize * (gridSize - 1) >= numberOfPlots:
            ncols -= 1
        
        return nrows, ncols



    def setupFileSelectionTab(self, tab):
        layout = QVBoxLayout(tab)  # Create a layout for this tab

        headerLabel = QLabel("SD Frequency and multi-plots\nshankar.dutt@anu.edu.au")
        headerLabel.setWordWrap(True)
        headerLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)  # Update alignment flag
        headerLabel.setFont(QFont('Arial', 18, QFont.Weight.Bold))  # Update font weight
        headerLabel.setStyleSheet("QLabel { color : white; margin-top: 20px; margin-bottom: 20px; }")

        # Update the stylesheet for QPushButton for PyQt6
        app.setStyleSheet("""
        QPushButton {
            background-color: #2b5b84;
            color: white;
            border-radius: 5px;
            padding: 10px;
        }
        QPushButton:hover {
            background-color: #1d4560;
        }
        """)

        layout.addWidget(headerLabel)

        selectFolderBtn = QPushButton("Select Folder")
        selectFolderBtn.clicked.connect(self.selectFolder)
        layout.addWidget(selectFolderBtn)

        self.fileListWidget = QListWidget()
        self.fileListWidget.setSelectionMode(QListWidget.SelectionMode.MultiSelection)
        layout.addWidget(self.fileListWidget)

        self.includeSubfoldersCheckbox = QCheckBox("Include Subfolders")
        layout.addWidget(self.includeSubfoldersCheckbox)

        plotButton = QPushButton("Plot")
        plotButton.clicked.connect(self.plotSelectedFiles)
        layout.addWidget(plotButton)

        # Initialize plot options and add them to the layout
        self.plotOptions = {
            'Histogram of dI': QCheckBox('Histogram of dI'),
            'Histogram of dt': QCheckBox('Histogram of dt'),
            'Histogram of Area': QCheckBox('Histogram of Area'),
            'Scatter plot dI vs dt': QCheckBox('Scatter plot dI vs dt'),
            'Scatter plot area vs dt': QCheckBox('Scatter plot area vs dt'),
            'Histogram of dt (log)': QCheckBox('Histogram of dt (log)'),
            'Scatter plot dI vs dt (log)': QCheckBox('Scatter plot dI vs dt (log)'),
            'Frequency plot': QCheckBox('Frequency plot')
        }
        for option in self.plotOptions.values():
            layout.addWidget(option)

        syncZoomButton = QPushButton("Sync Zoom")
        syncZoomButton.clicked.connect(self.syncZoomAcrossPlots)
        layout.addWidget(syncZoomButton)
        self.selectedFilesOrder = []
        
        # Connect the selection changed signal
        self.fileListWidget.itemSelectionChanged.connect(self.updateSelectionOrder)


    def setupPlotOptionsTab(self, tab):
        layout = QVBoxLayout(tab)
        
        # Use QFormLayout for a neat alignment of labels and fields including buttons
        formLayout = QFormLayout()

        # Histogram Bins
        self.histogramBinsSpinBox = QSpinBox()
        self.histogramBinsSpinBox.setMinimum(1)
        self.histogramBinsSpinBox.setMaximum(1000)
        self.histogramBinsSpinBox.setValue(100)
        formLayout.addRow("Histogram Bins:", self.histogramBinsSpinBox)

        # Marker Size
        self.markerSizeSpinBox = QSpinBox()
        self.markerSizeSpinBox.setMinimum(1)
        self.markerSizeSpinBox.setMaximum(100)
        self.markerSizeSpinBox.setValue(3)
        formLayout.addRow("Marker Size:", self.markerSizeSpinBox)

        # Linear Fit
        self.linearFitCheckBox = QCheckBox("Enable Linear Fit in the Frequency Plots")
        formLayout.addRow(self.linearFitCheckBox)

        # Fit between two points
        self.fitBetweenPointsCheckBox = QCheckBox("Fit Between Two Points")
        formLayout.addRow(self.fitBetweenPointsCheckBox)

        # Start point for fit
        self.fitStartPointSpinBox = QDoubleSpinBox()
        self.fitStartPointSpinBox.setMinimum(-np.inf)
        self.fitStartPointSpinBox.setMaximum(np.inf)
        self.fitStartPointSpinBox.setValue(0)

        # End point for fit
        self.fitEndPointSpinBox = QDoubleSpinBox()
        self.fitEndPointSpinBox.setMinimum(-np.inf)
        self.fitEndPointSpinBox.setMaximum(np.inf)
        self.fitEndPointSpinBox.setValue(1)

        self.fitRangeLabel = QLabel("Fit Range:")
        self.fitRangeLayout = QHBoxLayout()
        self.fitRangeLayout.addWidget(self.fitRangeLabel)
        self.fitRangeLayout.addWidget(self.fitStartPointSpinBox)
        self.fitRangeLayout.addWidget(self.fitEndPointSpinBox)
        formLayout.addRow(self.fitRangeLayout)

        self.fitBetweenPointsCheckBox.stateChanged.connect(self.toggleFitRangeVisibility)
        self.toggleFitRangeVisibility(False)

        self.overlayPlotsCheckBox = QCheckBox("Overlay Plots")
        formLayout.addRow(self.overlayPlotsCheckBox)

        self.alphaSpin = QDoubleSpinBox()
        self.alphaSpin.setMinimum(0.0)
        self.alphaSpin.setMaximum(1.0)
        self.alphaSpin.setSingleStep(0.1)
        self.alphaSpin.setValue(0.5)  # Default alpha value
        formLayout.addRow("Alpha:", self.alphaSpin)

        # Save directory input and button
        self.saveDirectoryLineEdit = QLineEdit()
        self.saveDirectoryButton = QPushButton("Select Save Folder")
        self.saveDirectoryButton.clicked.connect(self.selectSaveFolder)
        saveFolderLayout = QHBoxLayout()
        saveFolderLayout.addWidget(self.saveDirectoryLineEdit)
        saveFolderLayout.addWidget(self.saveDirectoryButton)
        formLayout.addRow("Save Figures To:", saveFolderLayout)

        # Update plots button
        updatePlotsButton = QPushButton("Update Plots")
        updatePlotsButton.clicked.connect(self.updatePlots)
        formLayout.addRow("", updatePlotsButton)  # Add an empty label for alignment

        # Save figures button
        saveFiguresButton = QPushButton("Save Figures")
        saveFiguresButton.clicked.connect(self.saveAllFigures)
        formLayout.addRow("", saveFiguresButton)  # Add an empty label for alignment

        layout.addLayout(formLayout)

        # Adjust overall layout settings
        layout.setSpacing(10)  # Adjust the spacing as needed
        tab.setLayout(layout)

    def toggleFitRangeVisibility(self, visible):
        self.fitRangeLabel.setVisible(visible)
        self.fitStartPointSpinBox.setVisible(visible)
        self.fitEndPointSpinBox.setVisible(visible)


    def selectSaveFolder(self):
        folder = QFileDialog.getExistingDirectory(self, "Select Folder")
        if folder:
            self.saveDirectoryLineEdit.setText(folder)

    def saveAllFigures(self):
        saveFolder = self.saveDirectoryLineEdit.text().strip()
        if not saveFolder:
            QMessageBox.information(self, "Info", "Please select a folder to save the figures.")
            return

        for i in range(self.rightPanel.count()):
            tab = self.rightPanel.widget(i)
            if hasattr(tab, "property"):
                tabCanvases = tab.property("canvases")
                if tabCanvases:
                    tabFolder = os.path.join(saveFolder, self.rightPanel.tabText(i))
                    if not os.path.exists(tabFolder):
                        os.makedirs(tabFolder)
                    for j, canvas in enumerate(tabCanvases):
                        figure = canvas.figure
                        figure.savefig(os.path.join(tabFolder, f"figure_{j+1}.png"), dpi=300, bbox_inches='tight')
        QMessageBox.information(self, "Info", "Figures saved successfully.")


           
    def selectFolder(self):
        folderPath = QFileDialog.getExistingDirectory(self, "Select Folder")
        if folderPath:
            self.populateFileList(folderPath,self.includeSubfoldersCheckbox.isChecked() )

    def populateFileList(self, folderPath, includeSubfolders):
        self.fileListWidget.clear()
        if includeSubfolders:
            for root, dirs, files in os.walk(folderPath):
                for file in files:
                    if file.endswith('.dataset.npz'):
                        fullPath = os.path.join(root, file)
                        self.fileListWidget.addItem(fullPath)
        else:
            for file in os.listdir(folderPath):
                if file.endswith('.dataset.npz'):
                    fullPath = os.path.join(folderPath, file)
                    self.fileListWidget.addItem(fullPath)
    
    def updateSelectionOrder(self):
        # Get the current selection
        currentSelection = {self.fileListWidget.item(i).text() for i in range(self.fileListWidget.count()) if self.fileListWidget.item(i).isSelected()}
        
        # Filter out any deselected items
        self.selectedFilesOrder = [f for f in self.selectedFilesOrder if f in currentSelection]
        
        # Add new selections in order
        for i in range(self.fileListWidget.count()):
            itemText = self.fileListWidget.item(i).text()
            if self.fileListWidget.item(i).isSelected() and itemText not in self.selectedFilesOrder:
                self.selectedFilesOrder.append(itemText)
    

    def updatePlots(self):
        # Update the class attributes with the new values from the UI
        self.histogram_bins = self.histogramBinsSpinBox.value()
        self.scatter_marker_size = self.markerSizeSpinBox.value()

        # Re-plot all selected files with the new settings
        self.plotSelectedFiles()


    @staticmethod
    def create_joint_plot(self, fig, data_x, data_y, min_x, max_x, min_y, max_y, plotType, alpha):
        # Create grid layout for the plots
        gs = fig.add_gridspec(4, 4)
        
        # Main scatter plot with specified axis limits
        ax_scatter = fig.add_subplot(gs[1:4, 0:3])
        if data_x is not None and data_y is not None:
            ax_scatter.scatter(data_x, data_y, s=int(self.markerSizeSpinBox.value()), alpha=alpha)
        ax_scatter.set_xlim(min_x, max_x)
        ax_scatter.set_ylim(min_y, max_y)
        
        # Histogram for X distribution
        ax_histx = fig.add_subplot(gs[0, 0:3], sharex=ax_scatter)
        if data_x is not None:
            ax_histx.hist(data_x, bins=self.histogramBinsSpinBox.value())
        ax_histx.set_ylabel(' ')
        
        # Histogram for Y distribution
        ax_histy = fig.add_subplot(gs[1:4, 3], sharey=ax_scatter)
        if data_y is not None:
            ax_histy.hist(data_y, bins=self.histogramBinsSpinBox.value(), orientation='horizontal')
        ax_histy.set_xlabel(' ')
        
        # Hide tick labels for histograms to avoid overlap
        plt.setp(ax_histx.get_xticklabels(), visible=False)
        plt.setp(ax_histy.get_yticklabels(), visible=False)

        ax_scatter.xaxis.set_major_formatter(ScalarFormatter(useMathText=True))
        ax_scatter.yaxis.set_major_formatter(ScalarFormatter(useMathText=True))
        if 'dI vs dt' in plotType and 'log' not in plotType:
            ax_scatter.set_xlabel('Δt (ms)')
            ax_scatter.set_ylabel('ΔI (nA)')
        elif 'area vs dt' in plotType:
            ax_scatter.set_xlabel('Δt (ms)')
            ax_scatter.set_ylabel('Area (C)')
        elif 'dI vs dt (log)' in plotType:
            ax_scatter.set_xlabel('Δt (ms) [log]')
            ax_scatter.set_ylabel('ΔI (nA)')

        ax_scatter.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
        ax_histx.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
        ax_histy.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
        fig.tight_layout()

        # Return the axes dictionary
        return {'ax_scatter': ax_scatter, 'ax_histx': ax_histx, 'ax_histy': ax_histy}
    
    def plotSelectedFiles(self):
        '''selectedFiles = [self.fileListWidget.item(i).text() for i in range(self.fileListWidget.count()) if self.fileListWidget.item(i).isSelected()]
        plotOptions = {key: option.isChecked() for key, option in self.plotOptions.items() if option.isChecked()}'''

        plotOptions = {key: option.isChecked() for key, option in self.plotOptions.items() if option.isChecked()}
        self.rightPanel.clear()

        self.rightPanel.clear()

        if not self.selectedFilesOrder or not plotOptions:
            return

        data = self.prepareData(self.selectedFilesOrder, plotOptions)

        for plotType, plotData in data.items():
            tab = QWidget()
            layout = QVBoxLayout(tab)
            plotLayout = QGridLayout()
            alpha = self.alphaSpin.value()
            tabCanvases = []

            if self.overlayPlotsCheckBox.isChecked():
                # Create a single figure and canvas for overlaid plots
                fig = Figure(figsize=(10, 5))
                canvas = FigureCanvas(fig)
                tabCanvases.append(canvas)

                min_x, max_x, min_y, max_y = self.calculateGlobalAxisLimits(plotData, plotType)

                if 'Scatter plot dI vs dt' in plotType and 'log' not in plotType:
                    axes_dict = self.create_joint_plot(self, fig, None, None, min_x, max_x, min_y, max_y, 'dI vs dt', alpha)
                    ax_scatter = axes_dict['ax_scatter']
                    ax_histx = axes_dict['ax_histx']
                    ax_histy = axes_dict['ax_histy']

                    for filePath, values in plotData.items():
                        alpha = self.alphaSpin.value()
                        label = os.path.basename(filePath)

                        ax_scatter.scatter(values[0], values[1], s=int(self.markerSizeSpinBox.value()), alpha=alpha, label=label)
                        ax_histx.hist(values[0], bins=self.histogramBinsSpinBox.value(), alpha=alpha)
                        ax_histy.hist(values[1], bins=self.histogramBinsSpinBox.value(), orientation='horizontal', alpha=alpha)
                    fig.tight_layout()
                    ax_scatter.legend()

                elif 'Scatter plot area vs dt' in plotType:
                    axes_dict = self.create_joint_plot(self, fig, None, None, min_x, max_x, min_y, max_y, 'area vs dt', alpha)
                    ax_scatter = axes_dict['ax_scatter']
                    ax_histx = axes_dict['ax_histx']
                    ax_histy = axes_dict['ax_histy']

                    for filePath, values in plotData.items():
                        alpha = self.alphaSpin.value()
                        label = os.path.basename(filePath)

                        ax_scatter.scatter(values[0], values[1], s=int(self.markerSizeSpinBox.value()), alpha=alpha, label=label)
                        ax_histx.hist(values[0], bins=self.histogramBinsSpinBox.value(), alpha=alpha)
                        ax_histy.hist(values[1], bins=self.histogramBinsSpinBox.value(), orientation='horizontal', alpha=alpha)
                    fig.tight_layout()
                    ax_scatter.legend()

                elif 'Scatter plot dI vs dt (log)' in plotType:
                    axes_dict = self.create_joint_plot(self, fig, None, None, np.log(min_x), np.log(max_x), min_y, max_y, 'dI vs dt (log)', alpha)
                    ax_scatter = axes_dict['ax_scatter']
                    ax_histx = axes_dict['ax_histx']
                    ax_histy = axes_dict['ax_histy']

                    for filePath, values in plotData.items():
                        alpha = self.alphaSpin.value()
                        label = os.path.basename(filePath)

                        ax_scatter.scatter(np.log(values[0]), values[1], s=int(self.markerSizeSpinBox.value()), alpha=alpha, label=label)
                        ax_histx.hist(np.log(values[0]), bins=self.histogramBinsSpinBox.value(), alpha=alpha)
                        ax_histy.hist(values[1], bins=self.histogramBinsSpinBox.value(), orientation='horizontal', alpha=alpha)
                    fig.tight_layout()
                    ax_scatter.legend()
                    

                elif 'Histogram of dt (log)' in plotType:
                    ax = fig.add_subplot(111)

                    for filePath, values in plotData.items():
                        alpha = self.alphaSpin.value()
                        label = os.path.basename(filePath)

                        log_values = np.log(values[values > 0])
                        ax.hist(log_values, bins=self.histogramBinsSpinBox.value(), range=(np.log(min_x), np.log(max_x)), alpha=alpha, label=label)
                    fig.tight_layout()
                    ax.legend()

                elif 'Frequency plot' in plotType:
                    ax = fig.add_subplot(111)

                    for filePath, values in plotData.items():
                        alpha = self.alphaSpin.value()
                        label = os.path.basename(filePath)

                        ax.plot(values[0], values[1], 'o', markersize=self.scatter_marker_size, alpha=alpha, label=label)

                        if self.linearFitCheckBox.isChecked():
                            if self.fitBetweenPointsCheckBox.isChecked():
                                start_point = self.fitStartPointSpinBox.value()
                                end_point = self.fitEndPointSpinBox.value()
                                mask = (values[0] >= start_point) & (values[0] <= end_point)
                                x_fit = values[0][mask]
                                y_fit = values[1][mask]
                                slope, intercept = np.polyfit(x_fit, y_fit, 1)
                            else:
                                slope, intercept = np.polyfit(values[0], values[1], 1)
                                x_fit = np.array([min_x, max_x])
                                y_fit = slope * x_fit + intercept

                            ax.plot(x_fit, slope * x_fit + intercept, '--', alpha=alpha)

                    ax.legend()

                else:  # Regular histogram
                    ax = fig.add_subplot(111)

                    for filePath, values in plotData.items():
                        alpha = self.alphaSpin.value()
                        label = os.path.basename(filePath)

                        ax.hist(values, bins=self.histogramBinsSpinBox.value(), range=(min_x, max_x), alpha=alpha, label=label)

                    ax.legend()

                fig.suptitle("Overlaid Plots", fontsize=12)
                fig.tight_layout()
                fig.tight_layout()
                fig.tight_layout()
                fig.tight_layout()
                fig.tight_layout()
                plotLayout.addWidget(canvas, 0, 0)
                

                
            else:
                row = col = 0
                max_col = 1
                
                min_x, max_x, min_y, max_y = self.calculateGlobalAxisLimits(plotData, plotType)

                for filePath, values in plotData.items():
                    fig = Figure(figsize=(5, 5))
                    canvas = FigureCanvas(fig)

                    if 'Scatter plot dI vs dt' in plotType and 'log' not in plotType:
                        axes_dict = self.create_joint_plot(self, fig, values[0], values[1], min_x, max_x, min_y, max_y, 'dI vs dt', alpha=1.0)
                        canvas.axes_dict = axes_dict
                        canvas.plotType = 'Scatter plot dI vs dt'
                        title = f"{os.path.basename(filePath)}\nEvents: {len(values[0])}"

                    elif 'Scatter plot area vs dt' in plotType:
                        axes_dict = self.create_joint_plot(self, fig, values[0], values[1], min_x, max_x, min_y, max_y, 'area vs dt', alpha=1.0)
                        canvas.axes_dict = axes_dict
                        canvas.plotType = 'Scatter plot area vs dt'
                        title = f"{os.path.basename(filePath)}\nEvents: {len(values[0])}"

                    elif 'Scatter plot dI vs dt (log)' in plotType:
                        axes_dict = self.create_joint_plot(self, fig, np.log(values[0]), values[1], np.log(min_x), np.log(max_x), min_y, max_y, 'dI vs dt (log)', alpha=1.0)
                        canvas.axes_dict = axes_dict
                        canvas.plotType = 'Scatter plot dI vs dt (log)'
                        title = f"{os.path.basename(filePath)}\nEvents: {len(values[0])}"

                    elif 'Histogram of dt (log)' in plotType:
                        log_values = np.log(values[values > 0])
                        axes_dict = self.create_histogram(fig, log_values, np.log(min_x), np.log(max_x), 'dt (log)', alpha)
                        canvas.axes_dict = axes_dict
                        canvas.plotType = 'Histogram of dt (log)'
                        title = f"{os.path.basename(filePath)}\nEvents: {len(log_values)}"

                    elif 'Frequency plot' in plotType:
                        self.create_frequency_plot(fig, values[0], values[1], min_x, max_x, min_y, max_y, alpha=1.0)
                        canvas.plotType = 'Frequency plot'
                        title = f"{os.path.basename(filePath)}\nEvents: {len(values[0])}"
                        if self.linearFitCheckBox.isChecked():
                            if self.fitBetweenPointsCheckBox.isChecked():
                                start_point = self.fitStartPointSpinBox.value()
                                end_point = self.fitEndPointSpinBox.value()
                                mask = (values[0] >= start_point) & (values[0] <= end_point)
                                x_fit = values[0][mask]
                                y_fit = values[1][mask]
                                slope, intercept = np.polyfit(x_fit, y_fit, 1)
                            else:
                                slope, intercept = np.polyfit(values[0], values[1], 1)
                                x_fit = np.array([min_x, max_x])
                                y_fit = slope * x_fit + intercept

                            ax = fig.axes[0]
                            ax.plot(x_fit, slope * x_fit + intercept, 'r--', label=f'Fit: {slope:.2f} Events/s, Intercept: {intercept:.2f}')
                            ax.legend()
                            title += f"\nFit: {slope:.2f} Events/s, Intercept: {intercept:.2f}"

                    else:
                        # Regular histogram
                        axes_dict = self.create_histogram(fig, values, min_x, max_x, plotType, alpha=1.0)
                        canvas.axes_dict = axes_dict
                        canvas.plotType = plotType
                        title = f"{os.path.basename(filePath)}\nEvents: {len(values)}"

                    fig.suptitle(title, fontsize=8)
                    
                    plotLayout.addWidget(canvas, row, col)
                    tabCanvases.append(canvas)

                    col += 1
                    if col > max_col:
                        col = 0
                        row += 1
                    
                    # Adjust layout
                    fig.tight_layout()

            layout.addLayout(plotLayout)

            if tabCanvases:
                toolbar = NavigationToolbar(tabCanvases[0], self)
                layout.addWidget(toolbar)

            fig.tight_layout()

            resetButton = QPushButton("Reset Zoom")
            resetButton.clicked.connect(self.resetZoom)
            layout.addWidget(resetButton)

            self.rightPanel.addTab(tab, plotType)
            tab.setProperty("canvases", tabCanvases)


    def create_frequency_plot(self, fig, times, event_numbers, min_x, max_x, min_y, max_y, alpha):
        ax = fig.add_subplot(111)
        ax.plot(times, event_numbers, 'o', markersize=self.scatter_marker_size, alpha = alpha)
        ax.set_xlim(min_x, max_x)
        #ax.set_ylim(min_y, max_y)
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Event Number')
        ax.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
        fig.tight_layout()
    

    def syncZoomAcrossPlots(self):
        currentTab = self.rightPanel.currentWidget()
        if hasattr(currentTab, "property"):
            tabCanvases = currentTab.property("canvases")
            if tabCanvases and len(tabCanvases) > 0:
                # Assuming the first canvas' figure's first axes as the reference
                referenceAxes = tabCanvases[0].figure.axes[0]
                xlim = referenceAxes.get_xlim()
                ylim = referenceAxes.get_ylim()
                
                for canvas in tabCanvases:
                    plotType = getattr(canvas, 'plotType', '')
                    if 'Scatter plot' in plotType:
                        if hasattr(canvas, 'axes_dict'):
                            canvas.axes_dict['ax_scatter'].set_xlim(xlim)
                            canvas.axes_dict['ax_scatter'].set_ylim(ylim)
                            canvas.axes_dict['ax_histx'].set_xlim(xlim)
                            canvas.axes_dict['ax_histy'].set_ylim(ylim)  # Reverse y-limits for horizontal histogram
                    elif 'Histogram' in plotType or 'Frequency plot' in plotType:
                        # For histograms and frequency plots, synchronize only the X-axis
                        ax = canvas.figure.axes[0]
                        ax.set_xlim(xlim)
                    else:
                        # For other plots, synchronize both axes
                        for ax in canvas.figure.axes:
                            ax.set_xlim(xlim)
                            ax.set_ylim(ylim)

                    canvas.draw_idle()

  

    def create_histogram(self, fig, values, min_x, max_x, plotType, alpha):
        ax = fig.add_subplot(111)
        ax.hist(values, bins=self.histogramBinsSpinBox.value(), range=(min_x, max_x),  alpha = alpha)
        ax.xaxis.set_major_formatter(ScalarFormatter(useMathText=True))
        if 'dI' in plotType:
            ax.set_xlabel('ΔI (nA)')
        elif 'dt' in plotType and 'log' not in plotType:
            ax.set_xlabel('Δt (s)')
        elif 'dt (log)' in plotType:
            ax.set_xlabel('log(Δt (ms))')
        ax.set_ylabel('Count')
        ax.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
        fig.tight_layout()

        # Return the axis dictionary
        return {'ax_hist': ax}


    def resetZoom(self):
        currentTab = self.rightPanel.currentWidget()
        if currentTab is None:
            return

        plotType = self.rightPanel.tabText(self.rightPanel.indexOf(currentTab))

        if plotType in self.global_limits:
            min_x, max_x, min_y, max_y = self.global_limits[plotType]

            tabCanvases = currentTab.property("canvases")
            if tabCanvases:
                for canvas in tabCanvases:
                    if plotType == 'Scatter plot dI vs dt' or plotType == 'Scatter plot area vs dt' or plotType == 'Scatter plot dI vs dt (log)':
                        if hasattr(canvas, 'axes_dict'):
                            canvas.axes_dict['ax_scatter'].set_xlim(min_x, max_x)
                            canvas.axes_dict['ax_scatter'].set_ylim(min_y, max_y)
                            canvas.axes_dict['ax_histx'].set_xlim(min_x, max_x)
                            canvas.axes_dict['ax_histy'].set_ylim(min_y, max_y)
                    elif plotType == 'Frequency plot':
                        ax = canvas.figure.axes[0]
                        ax.set_xlim(min_x, max_x)
                        #ax.set_ylim(min_y, max_y)
                    elif plotType in ['Histogram of dI', 'Histogram of dt', 'Histogram of Area', 'Histogram of dt (log)']:
                        ax = canvas.figure.axes[0]
                        ax.set_xlim(min_x, max_x)
                    else:
                        print(f"Unsupported plot type: {plotType}")

                    canvas.draw_idle()

    
    def prepareData(self, selectedFiles, plotOptions):
        data = {}
        for filePath in selectedFiles:
            fileData = np.load(filePath)
            X = fileData['X']
            
            for option in plotOptions:
                if option not in data:
                    data[option] = {}
                
                if 'Histogram' in option:
                    index = 0 if 'dI' in option else 3 if 'Area' in option else 4  # Assuming the order dI, dt, Area
                    values = X[:, index].astype(float)
                    # Remove NaN values
                    values = values[~np.isnan(values)]
                    data[option][filePath] = values
                    #print(f"Data type for {option}: {type(values)}")
                elif 'Histogram of dt (log)' in option:
                    # Assuming dt values are in the second column (index 1)
                    dt_values = fileData['X'][:, 4]
                    # Filter out non-positive values to avoid log(0) and log(negative)
                    positive_dt_values = dt_values[dt_values > 0]
                    # Apply logarithmic transformation
                    log_dt_values = np.log(positive_dt_values*1e3)
                    # Remove NaN values
                    log_dt_values = log_dt_values[~np.isnan(log_dt_values)]
                    data['Histogram of dt (log)'][filePath] = log_dt_values.astype(float)
                    #print(f"Data type for Histogram of dt (log): {type(log_dt_values)}")
                elif 'Frequency plot' in option:
                    event_numbers = np.arange(1, len(fileData['X']) + 1).astype(float)
                    times = fileData['X'][:, -1].astype(float)
                    # Remove NaN values
                    valid_indices = ~np.isnan(times)
                    event_numbers = event_numbers[valid_indices]
                    times = times[valid_indices]
                    data[option][filePath] = (times, event_numbers)
                    #print(f"Data type for Frequency plot: {type(times)}, {type(event_numbers)}")
                else:  # Scatter plots
                    x_index, y_index = (4, 0) if 'dI vs dt' in option else (4, 3)  # Assuming dI vs dt, Area vs dt
                    x_values = X[:, x_index]*1e3
                    y_values = X[:, y_index]
                    # Remove NaN values
                    valid_indices = ~np.isnan(x_values) & ~np.isnan(y_values)
                    x_values = x_values[valid_indices]
                    y_values = y_values[valid_indices]
                    data[option][filePath] = (x_values, y_values)
                    #print(f"Data type for {option}: {type(x_values)}, {type(y_values)}")
                    
        return data
    
    def calculateGlobalAxisLimits(self, plotData, plotType):
        min_x = min_y = np.inf
        max_x = max_y = -np.inf
        for values in plotData.values():
            if isinstance(values, dict):
                # If values is a dictionary, iterate over its values
                for subvalues in values.values():
                    if 'Histogram' in plotType:
                        curr_min_x = np.min(subvalues)
                        curr_max_x = np.max(subvalues)
                        #print(f"curr_min_x: {curr_min_x}, curr_max_x: {curr_max_x}")
                        if not np.isfinite(curr_min_x) or not np.isfinite(curr_max_x):
                            continue  # Skip this file if min_x or max_x is inf or -inf
                        if curr_min_x < curr_max_x:
                            min_x = min(min_x, curr_min_x)
                            max_x = max(max_x, curr_max_x)
                        else:
                            min_x = min(min_x, curr_min_x)
                            max_x = max(max_x, curr_min_x + 1e-6)  # Adjust max_x to be slightly larger than min_x
                    elif 'Frequency plot' in plotType:
                        x, y = subvalues
                        min_x = min(min_x, np.min(x))
                        max_x = max(max_x, np.max(x))
                        curr_min_y = np.min(y)
                        curr_max_y = np.max(y)
                        if not np.isfinite(curr_min_y) or not np.isfinite(curr_max_y):
                            continue  # Skip this file if min_y or max_y is inf or -inf
                        min_y = min(min_y, curr_min_y)
                        max_y = max(max_y, curr_max_y)
                    else:  # Scatter plots
                        x, y = subvalues
                        min_x = min(min_x, np.min(x))
                        max_x = max(max_x, np.max(x))
                        curr_min_y = np.min(y)
                        curr_max_y = np.max(y)
                        if not np.isfinite(curr_min_y) or not np.isfinite(curr_max_y):
                            continue  # Skip this file if min_y or max_y is inf or -inf
                        min_y = min(min_y, curr_min_y)
                        max_y = max(max_y, curr_max_y)
            else:
                if 'Histogram' in plotType:
                    curr_min_x = np.min(values)
                    curr_max_x = np.max(values)
                    #print(f"curr_min_x: {curr_min_x}, curr_max_x: {curr_max_x}")
                    if not np.isfinite(curr_min_x) or not np.isfinite(curr_max_x):
                        continue  # Skip this file if min_x or max_x is inf or -inf
                    if curr_min_x < curr_max_x:
                        min_x = min(min_x, curr_min_x)
                        max_x = max(max_x, curr_max_x)
                    else:
                        min_x = min(min_x, curr_min_x)
                        max_x = max(max_x, curr_min_x + 1e-6)  # Adjust max_x to be slightly larger than min_x
                elif 'Frequency plot' in plotType:
                    x, y = values
                    min_x = min(min_x, np.min(x))
                    max_x = max(max_x, np.max(x))
                    curr_min_y = np.min(y)
                    curr_max_y = np.max(y)
                    if not np.isfinite(curr_min_y) or not np.isfinite(curr_max_y):
                        continue  # Skip this file if min_y or max_y is inf or -inf
                    min_y = min(min_y, curr_min_y)
                    max_y = max(max_y, curr_max_y)
                else:  # Scatter plots
                    x, y = values
                    min_x = min(min_x, np.min(x))
                    max_x = max(max_x, np.max(x))
                    curr_min_y = np.min(y)
                    curr_max_y = np.max(y)
                    if not np.isfinite(curr_min_y) or not np.isfinite(curr_max_y):
                        continue  # Skip this file if min_y or max_y is inf or -inf
                    min_y = min(min_y, curr_min_y)
                    max_y = max(max_y, curr_max_y)

        if not np.isfinite(min_x) or not np.isfinite(max_x):
            min_x, max_x = 0, 1  # Set default values if min_x or max_x is inf or -inf
        if not np.isfinite(min_y) or not np.isfinite(max_y):
            min_y, max_y = 0, 1  # Set default values if min_y or max_y is inf or -inf

        # Store the global limits in the dictionary
        self.global_limits[plotType] = (min_x, max_x, min_y, max_y)

        return min_x, max_x, min_y, max_y
    

if __name__ == "__main__":

    app = QApplication(sys.argv)
    app.setStyle(QStyleFactory.create('Fusion'))  # Use Fusion or other available styles

    # Customize the palette for a darker, more modern look
    palette = QPalette()
    palette.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.WindowText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Base, QColor(25, 25, 25))
    palette.setColor(QPalette.ColorRole.AlternateBase, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ToolTipBase, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.ToolTipText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Text, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ButtonText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.BrightText, Qt.GlobalColor.red)
    palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.HighlightedText, Qt.GlobalColor.black)
    app.setPalette(palette)
    
    mainWindow = PlottingApplication()
    mainWindow.showMaximized()
    sys.exit(app.exec())



import sys
from PySide6.QtWidgets import (QApplication, QWidget, QVBoxLayout, QHBoxLayout,
                             QPushButton, QGroupBox, QLabel, QFileDialog,
                             QScrollArea, QMainWindow, QSizePolicy, QSplitter,
                             QRadioButton, QVBoxLayout, QLineEdit, QCheckBox, QDoubleSpinBox, QTabWidget,
                             QTableWidget, QTableWidgetItem, QSpinBox, QStyleFactory)
from PySide6.QtCore import Qt, QTimer, QSize
from PySide6.QtGui import QFont, QPalette, QColor
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
from matplotlib.figure import Figure

class MplCanvas(FigureCanvas):
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)

class ScatterCanvas(FigureCanvas):
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super(ScatterCanvas, self).__init__(fig)

class SDAnalysisApp(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle('SD Event Analysis App')
        self.setGeometry(100, 100, 1200, 800)

        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)
        self.main_splitter = QSplitter(Qt.Orientation.Horizontal)

        # Setup for the left panel (event categories and classifications)
        self.setup_left_panel()

        # Setup for the right panel (histograms, event plots, and event information)
        self.setup_right_panel()

        # Final layout setup
        self.central_layout = QHBoxLayout(self.central_widget)
        self.central_layout.addWidget(self.main_splitter)

        self.data = None
        self.events_data = {}
        self.classification_to_event_ids = {} 
        self.selected_event_ids = set() 


    def setup_left_panel(self):
        self.top_group = QGroupBox()
        self.top_group_layout = QVBoxLayout(self.top_group)
        
        # Setup components of top_group
        self.configure_top_group()

        # Setup scroll areas for Event Categories and Classifications
        self.event_categories_group_box, self.event_categories_layout = self.setup_scroll_area("Event Categories")
        self.event_classification_group_box, self.event_classification_layout = self.setup_scroll_area("Event Classifications")

        # Adjustments to place the Select All button at the bottom
        self.select_all_button = QPushButton("Select All")
        self.select_all_button.clicked.connect(self.select_all_classifications)

        # Left Splitter Configuration
        self.left_splitter = QSplitter(Qt.Orientation.Vertical)
        self.left_splitter.addWidget(self.top_group)
        self.left_splitter.addWidget(self.event_categories_group_box)
        self.left_splitter.addWidget(self.event_classification_group_box)
        self.left_splitter.addWidget(self.select_all_button)  # Add the Select All button here
        self.left_splitter.setSizes([10, 300, 300, 50])  # Adjust sizes accordingly
        self.main_splitter.addWidget(self.left_splitter)

    def configure_top_group(self):
        self.title_label = QLabel('SD Event Analysis App')
        self.title_label.setFont(QFont('Arial', 23, QFont.Weight.Bold))
        self.title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.subtitle_label = QLabel('shankar.dutt@anu.edu.au')
        self.subtitle_label.setFont(QFont('Arial', 15, QFont.Weight.Bold))
        self.subtitle_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.top_group_layout.addWidget(self.title_label)
        self.top_group_layout.addWidget(self.subtitle_label)

        self.file_button = QPushButton('Select File')
        self.file_button.clicked.connect(self.load_file)
        self.top_group_layout.addWidget(self.file_button)

        # Container for Similarity Threshold Label and SpinBox
        self.threshold_container = QWidget()
        self.threshold_container_layout = QHBoxLayout(self.threshold_container)

        self.threshold_label = QLabel("Similarity Threshold: ")
        self.threshold_container_layout.addWidget(self.threshold_label)
        
        self.threshold_input = QDoubleSpinBox()
        self.threshold_input.setSuffix(" %")
        self.threshold_input.setRange(0, 100)
        self.threshold_input.setValue(85)
        self.threshold_input.setSingleStep(1)
        
        # Setup QTimer
        self.threshold_timer = QTimer(self)  # Create a QTimer instance
        self.threshold_timer.setSingleShot(True)  # Set the timer to single-shot mode
        self.threshold_timer.timeout.connect(self.on_threshold_changed)  # Connect the timer's timeout signal to on_threshold_changed
        self.threshold_input.valueChanged.connect(self.start_threshold_timer)  # Connect valueChanged signal to a method that starts the timer
        
        self.threshold_container_layout.addWidget(self.threshold_input)

        self.reclassify_checkbox = QCheckBox("Reclassify the event categories based on threshold")
        self.reclassify_checkbox.stateChanged.connect(self.on_threshold_changed)
        self.top_group_layout.addWidget(self.threshold_container)
        self.top_group_layout.addWidget(self.reclassify_checkbox)

    def start_threshold_timer(self):
        self.threshold_timer.start(500)  # Start/restart the timer with a 500ms delay

    def setup_right_panel(self):
        self.right_panel = QWidget()
        self.right_splitter = QSplitter(Qt.Orientation.Vertical)

        # Create tabs for histograms and scatter plots
        self.tabs = QTabWidget()
        self.histograms_tab = QWidget()
        self.scatter_plots_tab = QWidget()
        self.tabs.addTab(self.histograms_tab, "Histograms")
        self.tabs.addTab(self.scatter_plots_tab, "Scatter Plots")

        # Setup histograms tab
        self.setup_histograms_tab()

        # Setup scatter plots tab
        self.setup_scatter_plots_tab()

        self.right_splitter.addWidget(self.tabs)

        # Histograms group setup
        self.histograms_group = QGroupBox("Histograms")
        self.histograms_horizontal_layout = QHBoxLayout(self.histograms_group)
        
        # Create and setup the first histogram canvas and toolbar for all events
        # self.all_events_layout = QVBoxLayout()
        # self.all_events_histogram_canvas = FigureCanvas(Figure(figsize=(5, 3)))
        # self.all_events_layout.addWidget(self.all_events_histogram_canvas)
        # self.all_events_histogram_toolbar = NavigationToolbar2QT(self.all_events_histogram_canvas, self.histograms_group)
        # self.all_events_histogram_toolbar.setIconSize(QSize(16, 16))
        # self.all_events_layout.addWidget(self.all_events_histogram_toolbar)
        # self.histograms_horizontal_layout.addLayout(self.all_events_layout)
        
        # # Create and setup the second histogram canvas and toolbar for selected classifications
        # self.selected_classifications_layout = QVBoxLayout()
        # self.selected_classifications_histogram_canvas = FigureCanvas(Figure(figsize=(5, 3)))
        # self.selected_classifications_layout.addWidget(self.selected_classifications_histogram_canvas)
        # self.selected_classifications_histogram_toolbar = NavigationToolbar2QT(self.selected_classifications_histogram_canvas, self.histograms_group)
        # self.selected_classifications_histogram_toolbar.setIconSize(QSize(16, 16))
        # self.selected_classifications_layout.addWidget(self.selected_classifications_histogram_toolbar)
        # self.histograms_horizontal_layout.addLayout(self.selected_classifications_layout)
        
        # self.right_splitter.addWidget(self.histograms_group)

        # Adjusted setup for event plots and event information to be side by side
        self.bottom_right_splitter = QSplitter(Qt.Orientation.Horizontal)

        # Event plots group setup
        self.event_plots_group = QGroupBox("Event Plots")
        self.event_plots_layout = QVBoxLayout(self.event_plots_group)
        self.event_plot_canvas = FigureCanvas(Figure(figsize=(5, 3)))
        self.event_plots_layout.addWidget(self.event_plot_canvas)
        self.event_plot_toolbar = NavigationToolbar2QT(self.event_plot_canvas, self.event_plots_group)
        self.event_plot_toolbar.setIconSize(QSize(16, 16))
        self.event_plots_layout.addWidget(self.event_plot_toolbar)

        # Navigation controls under the event plots
        self.event_navigation_layout = QHBoxLayout()
        self.prev_button = QPushButton("Previous")
        self.event_navigation_layout.addWidget(self.prev_button)
        # self.jump_to_label = QLabel("Jump to:")
        # self.event_navigation_layout.addWidget(self.jump_to_label)
        # self.jump_spinbox = QSpinBox()
        # self.jump_spinbox.setMinimum(1)  # Adjust as needed
        # self.jump_spinbox.setMaximum(100)  # Adjust based on your events count
        # self.event_navigation_layout.addWidget(self.jump_spinbox)
        # self.jump_button = QPushButton("Jump")
        # self.event_navigation_layout.addWidget(self.jump_button)
        self.next_button = QPushButton("Next")
        self.event_navigation_layout.addWidget(self.next_button)
        self.event_plots_layout.addLayout(self.event_navigation_layout)
        self.next_button.clicked.connect(self.next_event)  # Connect to the method handling the next event action
        self.prev_button.clicked.connect(self.previous_event)  # Connect to the method handling the previous event action


        self.bottom_right_splitter.addWidget(self.event_plots_group)

        # Event information group setup
        self.table_group = QGroupBox("Event Information")
        self.table_layout = QVBoxLayout(self.table_group)
        self.event_info_table = QTableWidget(10, 3)  # Adjust row, column count as needed
        self.event_info_table.setHorizontalHeaderLabels(['Type', 'Value', 'Description'])
        self.table_layout.addWidget(self.event_info_table)
        self.bottom_right_splitter.addWidget(self.table_group)

        self.right_splitter.addWidget(self.bottom_right_splitter)
        self.main_splitter.addWidget(self.right_splitter)

        # Setting initial sizes for splitters
        self.main_splitter.setSizes([300, 800])  # Adjust as needed
        self.right_splitter.setSizes([400, 350])  # Adjust as needed
        self.bottom_right_splitter.setSizes([400, 400])  # Adjust as needed

    def setup_histograms_tab(self):
        self.histograms_layout = QVBoxLayout(self.histograms_tab)
        self.histograms_group = QGroupBox("Histograms")
        self.histograms_horizontal_layout = QHBoxLayout(self.histograms_group)
        
        # Create and setup the first histogram canvas and toolbar for all events
        self.all_events_layout = QVBoxLayout()
        self.all_events_histogram_canvas = FigureCanvas(Figure(figsize=(5, 3)))
        self.all_events_layout.addWidget(self.all_events_histogram_canvas)
        self.all_events_histogram_toolbar = NavigationToolbar2QT(self.all_events_histogram_canvas, self.histograms_group)
        self.all_events_histogram_toolbar.setIconSize(QSize(16, 16))
        self.all_events_layout.addWidget(self.all_events_histogram_toolbar)
        self.histograms_horizontal_layout.addLayout(self.all_events_layout)
        
        # Create and setup the second histogram canvas and toolbar for selected classifications
        self.selected_classifications_layout = QVBoxLayout()
        self.selected_classifications_histogram_canvas = FigureCanvas(Figure(figsize=(5, 3)))
        self.selected_classifications_layout.addWidget(self.selected_classifications_histogram_canvas)
        self.selected_classifications_histogram_toolbar = NavigationToolbar2QT(self.selected_classifications_histogram_canvas, self.histograms_group)
        self.selected_classifications_histogram_toolbar.setIconSize(QSize(16, 16))
        self.selected_classifications_layout.addWidget(self.selected_classifications_histogram_toolbar)
        self.histograms_horizontal_layout.addLayout(self.selected_classifications_layout)
        
        self.histograms_layout.addWidget(self.histograms_group)

    def setup_scatter_plots_tab(self):
        self.scatter_plots_layout = QVBoxLayout(self.scatter_plots_tab)
        self.scatter_plots_group = QGroupBox("Scatter Plots")
        self.scatter_plots_horizontal_layout = QHBoxLayout(self.scatter_plots_group)

        # Create and setup the first scatter plot canvas and toolbar for all events
        self.all_events_scatter_layout = QVBoxLayout()
        self.all_events_scatter_canvas = ScatterCanvas(self, width=5, height=3)
        self.all_events_scatter_layout.addWidget(self.all_events_scatter_canvas)
        self.all_events_scatter_toolbar = NavigationToolbar2QT(self.all_events_scatter_canvas, self.scatter_plots_group)
        self.all_events_scatter_toolbar.setIconSize(QSize(16, 16))
        self.all_events_scatter_layout.addWidget(self.all_events_scatter_toolbar)
        self.scatter_plots_horizontal_layout.addLayout(self.all_events_scatter_layout)

        # Create and setup the second scatter plot canvas and toolbar for selected classifications
        self.selected_classifications_scatter_layout = QVBoxLayout()
        self.selected_classifications_scatter_canvas = ScatterCanvas(self, width=5, height=3)
        self.selected_classifications_scatter_layout.addWidget(self.selected_classifications_scatter_canvas)
        self.selected_classifications_scatter_toolbar = NavigationToolbar2QT(self.selected_classifications_scatter_canvas, self.scatter_plots_group)
        self.selected_classifications_scatter_toolbar.setIconSize(QSize(16, 16))
        self.selected_classifications_scatter_layout.addWidget(self.selected_classifications_scatter_toolbar)
        self.scatter_plots_horizontal_layout.addLayout(self.selected_classifications_scatter_layout)

        self.scatter_plots_layout.addWidget(self.scatter_plots_group)


    def setup_scroll_area(self, title):
        scroll_area = QScrollArea()  # Create the scroll area
        scroll_area.setWidgetResizable(True)
        
        container_widget = QWidget()  
        layout = QVBoxLayout(container_widget) 
        
        scroll_area.setWidget(container_widget) 

        group_box = QGroupBox(title)
        group_layout = QVBoxLayout(group_box)  
        group_layout.addWidget(scroll_area) 

        return group_box, layout 

    def on_threshold_changed(self):
        #current_value = self.threshold_input.value()
        self.prepare_and_display_event_data()  

    
    def load_file(self):
        file_name, _ = QFileDialog.getOpenFileName(self, 'Open NPZ File', '', 'NPZ Files (*event_fitting.npz)')
        if file_name:
            self.data = np.load(file_name, allow_pickle=True)  # Directly load and store data
            self.prepare_and_display_event_data()

    def prepare_and_display_event_data(self):
        if self.data is not None:
            self.events_data.clear()  # Ensure we're starting fresh
            self.clear_layout(self.event_categories_layout)
            self.clear_layout(self.event_classification_layout)
            self.plot_all_events_histogram()

            # Process and categorize events
            for key in self.data.files:
                if 'SEGMENT_INFO' in key and 'number_of_segments' in key:
                    event_id = int(key.split('_')[2])
                    # Extract mean_diffs to determine if reclassification changes segment count
                    mean_diffs_key = f'SEGMENT_INFO_{event_id}_segment_mean_diffs'
                    segment_widths_key = f'SEGMENT_INFO_{event_id}_segment_widths_time'
                    if mean_diffs_key in self.data and segment_widths_key in self.data:
                        mean_diffs = self.data[mean_diffs_key]
                        segment_widths = self.data[segment_widths_key]
                        _, new_segment_count = self.classify_event(mean_diffs)  # Use new segment count
                        self.events_data.setdefault(new_segment_count, []).append((event_id, mean_diffs, segment_widths))
                    else:
                        new_segment_count = int(self.data[key][0])
                        self.events_data.setdefault(new_segment_count, []).append((event_id, [], []))
            
            for num_segments in sorted(self.events_data):
                events_data = self.events_data[num_segments]
                event_ids = [event_data[0] for event_data in events_data]
                radio_button = QRadioButton(f"{num_segments} segments ({len(event_ids)} events)")
                radio_button.segment_number = num_segments
                radio_button.toggled.connect(self.on_radio_button_toggled)
                self.event_categories_layout.addWidget(radio_button)
            
            

    def on_radio_button_toggled(self):
        radio_button = self.sender()
        if radio_button.isChecked():
            self.update_classification_group(radio_button.segment_number)

    def update_classification_group(self, segment_number):
        self.clear_layout(self.event_classification_layout)
        classification_counts = {}
        self.classification_checkboxes = []
        self.classification_to_event_ids.clear()  # Reset the mapping

        for event_data in self.events_data.get(segment_number, []):
            event_id = event_data[0]
            mean_diffs_key = f'SEGMENT_INFO_{event_id}_segment_mean_diffs'
            if mean_diffs_key in self.data:
                mean_diffs = self.data[mean_diffs_key]
                classification, _ = self.classify_event(mean_diffs)
                classification_counts[classification] = classification_counts.get(classification, 0) + 1
                # Store event IDs under each classification
                if classification not in self.classification_to_event_ids:
                    self.classification_to_event_ids[classification] = []
                self.classification_to_event_ids[classification].append(event_id)

        for classification, count in sorted(classification_counts.items(), key=lambda x: x[0]):
            checkbox = QCheckBox(f"Category {classification} ({count} events)")
            # Now each checkbox represents a whole category/classification of events
            checkbox.classification = classification  # Use classification to identify the group
            checkbox.stateChanged.connect(self.on_checkbox_state_changed)
            self.event_classification_layout.addWidget(checkbox)
            self.classification_checkboxes.append(checkbox)

        # Ensure the "Select All" button works with the newly added checkboxes
        self.select_all_button.clicked.disconnect()  # First, disconnect any existing connection
        self.select_all_button.clicked.connect(self.select_all_classifications)  # Reconnect with the updated list of checkboxes



    def classify_event(self, mean_diffs):
        threshold_ratio = self.threshold_input.value() / 100.0

        # Initial classification
        if len(mean_diffs) <= 1:
            return "1", 1  # Single segment events are automatically classified as "1"

        # Initial classifications: each segment starts as unique
        classifications = ["1"]
        
        for i in range(1, len(mean_diffs)):
            found_similar = False
            for j in range(i):
                # Calculate the ratio to compare the mean differences
                ratio = mean_diffs[i] / mean_diffs[j] if mean_diffs[i] > mean_diffs[j] else mean_diffs[j] / mean_diffs[i]
                
                # If the ratio is within the similarity threshold
                if ratio <= (1 / threshold_ratio):
                    classifications.append(classifications[j])  # Assign the same classification as the similar segment
                    found_similar = True
                    break
            
            if not found_similar:
                # Assign a new unique classification
                max_classification = max([int(c) for c in classifications])
                classifications.append(str(max_classification + 1))

        # Check if reclassification based on threshold is checked
        if self.reclassify_checkbox.isChecked() and len(classifications) > 1:
            # Perform reclassification
            new_classifications = [classifications[0]]
            for i in range(1, len(classifications)):
                # Merge adjacent similar classifications
                if classifications[i] == classifications[i - 1]:
                    continue  # Skip adding it to new classifications
                new_classifications.append(classifications[i])
            classifications = new_classifications

        # Calculate new segment count based on reclassification
        new_segment_count = len(classifications)

        # Return both the classification string and the new segment count
        return ''.join(classifications), new_segment_count


    def clear_layout(self, layout):
        while layout.count():
            child = layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

    def select_all_classifications(self):
        try:
            if any(not checkbox.isChecked() for checkbox in self.classification_checkboxes):
                for checkbox in self.classification_checkboxes:
                    checkbox.setChecked(True)
            else:
                for checkbox in self.classification_checkboxes:
                    checkbox.setChecked(False)

            # Update the button text based on the new state
            self.select_all_button.setText("Unselect All" if any(checkbox.isChecked() for checkbox in self.classification_checkboxes) else "Select All")
        except:
            pass


    def plot_all_events_histogram(self):
        max_mean_diffs = []
        for key in self.data.files:
            if 'segment_mean_diffs' in key:
                mean_diffs = self.data[key]
                max_mean_diffs.append(np.max(mean_diffs))
        
        # Clear the figure/canvas before plotting
        self.all_events_histogram_canvas.figure.clear()

        ax = self.all_events_histogram_canvas.figure.subplots()
        
        # Calculate bin edges with 'auto' and then double the number of bins
        _, bins_auto = np.histogram(max_mean_diffs, bins='auto')
        num_bins_auto = len(bins_auto) - 1  # Number of bins is one less than the number of edges
        doubled_num_bins = num_bins_auto * 2  # Double the number of bins
        
        # Use linspace to create new bin edges with doubled number of bins
        new_bins = np.linspace(bins_auto[0], bins_auto[-1], doubled_num_bins + 1)
        
        ax.hist(max_mean_diffs, bins=new_bins)
        ax.set_title('Max Segment Mean Diffs for All Events')
        ax.set_xlabel('Max Mean Diff')
        ax.set_ylabel('Frequency')
        self.all_events_histogram_canvas.figure.tight_layout()
        self.all_events_histogram_canvas.draw()
        # Plot the corresponding scatter plot
        self.plot_all_events_scatter()

    def on_checkbox_state_changed(self):
        self.selected_event_ids.clear()  # Clear and repopulate based on current selections

        for checkbox in self.classification_checkboxes:
            if checkbox.isChecked():
                # Add all event IDs from this classification to the selected set
                event_ids = self.classification_to_event_ids.get(checkbox.classification, [])
                self.selected_event_ids.update(event_ids)
        
        # Plot and display segment info for the first selected event, if any
        self.current_event_index = 0
        if self.selected_event_ids:
            first_event_id = next(iter(sorted(self.selected_event_ids)))
            self.plot_event_data(first_event_id)
            self.display_segment_info(first_event_id)

        # Optionally update plots/information here
        self.plot_selected_events_histogram()
    
    def setup_navigation_buttons(self):
        self.current_event_index = 0  # Initialize the current index
        self.previous_button.clicked.connect(self.previous_event)
        self.next_button.clicked.connect(self.next_event)

    def previous_event(self):
        if self.selected_event_ids and self.current_event_index > 0:
            self.current_event_index -= 1
            event_id = sorted(self.selected_event_ids)[self.current_event_index]
            self.plot_event_data(event_id)
            self.display_segment_info(event_id)

    def next_event(self):
        if self.selected_event_ids and self.current_event_index < len(self.selected_event_ids) - 1:
            self.current_event_index += 1
            event_id = sorted(self.selected_event_ids)[self.current_event_index]
            self.plot_event_data(event_id)
            self.display_segment_info(event_id)

    def plot_selected_events_histogram(self):
        all_mean_diffs = []
        for event_id in self.selected_event_ids:
            key = f'SEGMENT_INFO_{event_id}_segment_mean_diffs'
            if key in self.data:
                mean_diffs = self.data[key]
                all_mean_diffs.extend(mean_diffs)              
        
        if len(all_mean_diffs)>1:
            # Clear the figure/canvas before plotting
            self.selected_classifications_histogram_canvas.figure.clear()

            ax = self.selected_classifications_histogram_canvas.figure.subplots()
            
            # Decide on the number of bins
            num_bins = 'auto'  # Default
            
            # Plot the histogram with bar boundaries
            ax.hist(all_mean_diffs, bins=num_bins, edgecolor='black')

            ax.set_title('Segment Mean Diffs for Selected Events')
            ax.set_xlabel('Mean Diff')
            ax.set_ylabel('Frequency')
            self.selected_classifications_histogram_canvas.figure.tight_layout()
            self.selected_classifications_histogram_canvas.draw()
        
        # Plot the corresponding scatter plot
        self.plot_selected_events_scatter()

    def plot_all_events_scatter(self):
        max_mean_diffs = []
        event_widths = []
        for key in self.data.files:
            if 'segment_mean_diffs' in key:
                mean_diffs = self.data[key]
                max_mean_diffs.append(np.max(mean_diffs))

            if 'event_width' in key:
                event_widths.append(self.data[key])

        if len(max_mean_diffs) > 0 and len(event_widths) > 0:
            # Clear the figure/canvas before plotting
            self.all_events_scatter_canvas.figure.clear()

            ax = self.all_events_scatter_canvas.figure.subplots()
            ax.scatter(np.log(np.array(event_widths)*1e3), max_mean_diffs)
            ax.set_title('Max Segment Mean Diffs vs log(Event Width) for All Events')
            ax.set_xlabel('log(Δt (ms))')
            ax.set_ylabel('ΔI')
            self.all_events_scatter_canvas.figure.tight_layout()
            self.all_events_scatter_canvas.draw()


    def plot_selected_events_scatter(self):
        all_mean_diffs = []
        all_segment_widths = []
        for event_id in self.selected_event_ids:
            for segment_count, events in self.events_data.items():
                for event_data in events:
                    if event_data[0] == event_id:
                        _, mean_diffs, segment_widths = event_data
                        all_mean_diffs.extend(mean_diffs)
                        all_segment_widths.extend(segment_widths)
                        break

        if len(all_mean_diffs) > 0 and len(all_segment_widths) > 0:
            # Clear the figure/canvas before plotting
            self.selected_classifications_scatter_canvas.figure.clear()

            ax = self.selected_classifications_scatter_canvas.figure.subplots()
            ax.scatter(np.log(np.array(all_segment_widths)*1e3), all_mean_diffs)
            ax.set_title('Segment Mean Diffs vs log(dt (ms)) for Selected Events')
            ax.set_xlabel('log(Δt (ms))')
            ax.set_ylabel('Mean Diff')
            self.selected_classifications_scatter_canvas.figure.tight_layout()
            self.selected_classifications_scatter_canvas.draw()

    def plot_event_data(self, event_id):
        # Retrieve event data
        x_values = self.data[f'EVENT_DATA_{event_id}_part_0']
        y_values_event = self.data[f'EVENT_DATA_{event_id}_part_1']
        y_values_fit = self.data[f'EVENT_DATA_{event_id}_part_3']
        
        # Clear the plot
        self.event_plot_canvas.figure.clear()
        
        # Create a new plot
        ax = self.event_plot_canvas.figure.subplots()
        ax.plot(x_values, y_values_event, label='Event Data')
        ax.plot(x_values, y_values_fit, label='Fit Data', linestyle='--')
        
        # Adding legends, title, and labels
        ax.legend()
        ax.set_title(f'Event {event_id} Data')
        ax.set_xlabel('Time')
        ax.set_ylabel('Data')
        
        self.event_plot_canvas.figure.tight_layout()
        # Refresh the canvas
        self.event_plot_canvas.draw()

    

    def display_segment_info(self, event_id):
        segment_info_keys = [
            f'SEGMENT_INFO_{event_id}_number_of_segments',
            f'SEGMENT_INFO_{event_id}_segment_mean_diffs',
            f'SEGMENT_INFO_{event_id}_segment_widths_time'
        ]
        
        # Extract segment info
        number_of_segments = self.data[segment_info_keys[0]][0]
        segment_mean_diffs = self.data[segment_info_keys[1]]
        segment_widths_time = self.data[segment_info_keys[2]]
        
        # Configure table for display
        self.event_info_table.setRowCount(int(number_of_segments))
        self.event_info_table.setColumnCount(3)
        self.event_info_table.setHorizontalHeaderLabels(['Segment', 'Mean Diff', 'Width Time'])
        
        # Populate the table and round the data
        for i in range(int(number_of_segments)):
            self.event_info_table.setItem(i, 0, QTableWidgetItem(f"{i + 1}"))
            self.event_info_table.setItem(i, 1, QTableWidgetItem(f"{segment_mean_diffs[i]:.3g}"))  # Rounded to 3 significant figures
            self.event_info_table.setItem(i, 2, QTableWidgetItem(f"{segment_widths_time[i]:.3g}"))  # Rounded to 3 significant figures

        # Adjust column widths to fit the content
        self.event_info_table.resizeColumnsToContents()

        # Optionally, you can also adjust the row heights if needed
        self.event_info_table.resizeRowsToContents()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    app.setStyle(QStyleFactory.create('Fusion'))  #
    palette = QPalette()
    palette.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.WindowText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Base, QColor(25, 25, 25))
    palette.setColor(QPalette.ColorRole.AlternateBase, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ToolTipBase, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.ToolTipText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Text, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ButtonText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.BrightText, Qt.GlobalColor.red)
    palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.HighlightedText, Qt.GlobalColor.black)
    app.setPalette(palette)
    mainWin = SDAnalysisApp()
    mainWin.showMaximized()
    sys.exit(app.exec())



import sys
from PySide6.QtWidgets import QApplication, QMainWindow, QWidget, QHBoxLayout, QVBoxLayout, QPushButton, QCheckBox, QListWidget, QLabel, QFileDialog, QLineEdit, QMessageBox, QListWidgetItem, QSplitter, QGroupBox, QFormLayout, QComboBox, QSpinBox,  QStyleFactory
from PySide6.QtCore import Qt
from PySide6.QtGui import QPalette, QColor
import os
import numpy as np
from neo.rawio import AxonRawIO
from scipy.ndimage import uniform_filter1d
import matplotlib
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.figure import Figure
import h5py
from scipy.signal import butter, bessel, cheby1, cheby2, ellip, firwin, lfilter, sosfilt, sosfilt_zi

matplotlib.use('Qt5Agg')

class MplCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        fig.tight_layout()
        super().__init__(fig)

def load_abf_file(file_path):
    # Create an instance of AxonRawIO
    raw_io = AxonRawIO(filename=file_path)

    # Parse the header information
    raw_io.parse_header()

    # Define the channel index you're interested in
    channel_index = 0

    # Read the signal size for the given block and segment without specifying channel_indexes
    signal_size = raw_io.get_signal_size(block_index=0, seg_index=0)

    # Now, read the analog signal data
    # Here, we specify channel_indexes when reading the chunk to ensure we only get data for the desired channel
    data = raw_io.get_analogsignal_chunk(block_index=0, seg_index=0, i_start=0, i_stop=signal_size, channel_indexes=[channel_index])

    # Convert the chunk to a physical quantity
    data = raw_io.rescale_signal_raw_to_float(data, dtype='float64', channel_indexes=[channel_index]).flatten()

    # Get the sampling rate for the specified channel
    sampling_rate = raw_io.get_signal_sampling_rate()

    # Convert the data to time
    time = np.arange(len(data)) / sampling_rate

    return data, sampling_rate, time

def load_hdf5_file(file_path):
    with h5py.File(file_path, 'r') as f:
        selected_data = f['selected_data'][()]
        sampling_rate = f.attrs['sampling_rate']
        time = np.arange(len(selected_data)) / sampling_rate
    return selected_data, sampling_rate, time

def calculate_rolling_stats(data, sampling_rate, avg_window_size_in_ms):
    avg_window_size_samples = int((avg_window_size_in_ms / 1000) * sampling_rate)
    rolling_avg = uniform_filter1d(data, size=avg_window_size_samples)
    return rolling_avg

class SegmentWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QHBoxLayout()
        self.setLayout(layout)

        self.start_label = QLabel("Start (s):")
        self.start_input = QLineEdit()
        self.end_label = QLabel("End (s):")
        self.end_input = QLineEdit()
        self.add_button = QPushButton("Add Segment")

        layout.addWidget(self.start_label)
        layout.addWidget(self.start_input)
        layout.addWidget(self.end_label)
        layout.addWidget(self.end_input)
        layout.addWidget(self.add_button)

class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("SD ABF File Plotter and Selector App")
        self.setGeometry(100, 100, 1200, 800)

        self.data = None
        self.sampling_rate = None
        self.time = None
        self.selected_regions = None
        self.segment_widgets = []
        self.segments = []

        main_splitter = QSplitter(Qt.Orientation.Horizontal)

        left_widget = QWidget()
        left_layout = QVBoxLayout()
        left_widget.setLayout(left_layout)

        right_widget = QWidget()
        right_layout = QVBoxLayout()
        right_widget.setLayout(right_layout)

        main_splitter.addWidget(left_widget)
        main_splitter.addWidget(right_widget)
        main_splitter.setSizes([300, 900])

        self.setCentralWidget(main_splitter)

        nth_element_layout = QHBoxLayout()

        # Left Side
        self.app_name_label = QLabel("SD ABF File Plotter and Selector App")
        self.app_name_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.app_name_label.setStyleSheet("font-size: 22px; font-weight: bold;")
        self.email_label = QLabel("shankar.dutt@anu.edu.au")
        self.email_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.select_folder_btn = QPushButton("Select Folder")
        self.select_folder_btn.clicked.connect(self.select_folder)

        self.include_subfolders_chk = QCheckBox("Include Subfolders")

        self.files_list_widget = QListWidget()
        self.files_list_widget.setSelectionMode(QListWidget.SelectionMode.MultiSelection)

        self.folder_path_label = QLabel(" ")
        self.folder_path_label.setWordWrap(True)

        self.nth_element_label = QLabel("Plot nth element:")
        self.nth_element_spinbox = QSpinBox()
        self.nth_element_spinbox.setMinimum(1)
        self.nth_element_spinbox.setMaximum(1000)
        self.nth_element_spinbox.setValue(1)

        nth_element_layout.addWidget(self.nth_element_label)
        nth_element_layout.addWidget(self.nth_element_spinbox)

        self.low_pass_filter_chk = QCheckBox("Apply Low Pass Filter")
        self.low_pass_filter_chk.stateChanged.connect(self.toggle_low_pass_filter)

        self.filter_type_label = QLabel("Filter Type:")
        self.filter_type_dropdown = QComboBox()
        self.filter_type_dropdown.addItems(["Butterworth", "Bessel", "Chebyshev I", "Chebyshev II", "Elliptic", "FIR", "IIR"])
        self.filter_type_dropdown.setEnabled(False)

        self.cutoff_frequency_label = QLabel("Cutoff Frequency (kHz):")
        self.cutoff_frequency_spinbox = QSpinBox()
        self.cutoff_frequency_spinbox.setRange(1, 1000)
        self.cutoff_frequency_spinbox.setValue(10)
        self.cutoff_frequency_spinbox.setEnabled(False)

        self.plot_btn = QPushButton("Plot Selected File")
        self.plot_btn.clicked.connect(self.plot_selected_file)

        left_layout.addWidget(self.app_name_label)
        left_layout.addWidget(self.email_label)
        left_layout.addWidget(self.select_folder_btn)
        left_layout.addWidget(self.include_subfolders_chk)
        left_layout.addWidget(self.files_list_widget)
        left_layout.addWidget(self.folder_path_label)
        left_layout.addLayout(nth_element_layout)
        left_layout.addWidget(self.low_pass_filter_chk)
        left_layout.addWidget(self.filter_type_label)
        left_layout.addWidget(self.filter_type_dropdown)
        left_layout.addWidget(self.cutoff_frequency_label)
        left_layout.addWidget(self.cutoff_frequency_spinbox)
        left_layout.addWidget(self.plot_btn)

        # Right Side
        right_splitter = QSplitter(Qt.Orientation.Vertical)
        right_layout.addWidget(right_splitter)

        top_right_widget = QWidget()
        top_right_layout = QVBoxLayout()
        top_right_widget.setLayout(top_right_layout)

        bottom_right_widget = QWidget()
        bottom_right_layout = QVBoxLayout()
        bottom_right_widget.setLayout(bottom_right_layout)

        right_splitter.addWidget(top_right_widget)
        right_splitter.addWidget(bottom_right_widget)

        # Top Right
        self.canvas = MplCanvas(self, width=5, height=4, dpi=100)
        top_right_layout.addWidget(self.canvas)

        self.mpl_toolbar = NavigationToolbar2QT(self.canvas, self)
        top_right_layout.addWidget(self.mpl_toolbar)

        # Bottom Right
        bottom_right_widget = QWidget()
        bottom_right_layout = QHBoxLayout()
        bottom_right_widget.setLayout(bottom_right_layout)
        right_splitter.addWidget(bottom_right_widget)

        self.calculate_avg_group = QGroupBox("Calculate and Plot Avg")
        self.calculate_avg_layout = QFormLayout()
        self.calculate_avg_group.setLayout(self.calculate_avg_layout)

        self.calculate_avg_chk = QCheckBox("Calculate and Plot Avg")
        self.calculate_avg_chk.stateChanged.connect(self.toggle_avg_window_size)

        self.avg_window_size_label = QLabel("Avg Window Size (ms):")
        self.avg_window_size_input = QLineEdit("10")
        #self.avg_window_size_input.setEnabled(False)
        self.avg_window_size_input.editingFinished.connect(self.update_plot)

        self.update_plot_btn = QPushButton("Update Plot")
        #self.update_plot_btn.setEnabled(False)
        self.update_plot_btn.clicked.connect(self.update_plot)

        self.calculate_avg_layout.addRow(self.calculate_avg_chk)
        self.calculate_avg_layout.addRow( self.avg_window_size_label, self.avg_window_size_input)
        self.calculate_avg_layout.addRow(self.update_plot_btn)

        self.threshold_group = QGroupBox("Select Regions within Threshold")
        self.threshold_layout = QFormLayout()
        self.threshold_group.setLayout(self.threshold_layout)

        self.threshold_chk = QCheckBox("Select Regions within Threshold")
        self.threshold_chk.stateChanged.connect(self.toggle_threshold_input)

        self.threshold_value_label = QLabel("Value for selecting regions(pA):")
        self.threshold_value_input = QSpinBox()
        self.threshold_value_input.setRange(0, 1000000)
        self.threshold_value_input.setValue(0)
        self.threshold_value_input.setEnabled(False)
        self.threshold_value_input.editingFinished.connect(self.update_plot)

        self.threshold_label = QLabel("Threshold (% of the value above):")
        self.threshold_input = QLineEdit("75")
        self.threshold_input.setEnabled(False)
        self.threshold_input.editingFinished.connect(self.update_plot)

        self.threshold_layout.addRow(self.threshold_chk)
        self.threshold_layout.addRow(self.threshold_value_label, self.threshold_value_input)
        self.threshold_layout.addRow( self.threshold_label, self.threshold_input)

        self.select_segments_group = QGroupBox("Select Segments")
        self.select_segments_layout = QVBoxLayout()
        self.select_segments_group.setLayout(self.select_segments_layout)

        self.select_segments_chk = QCheckBox("Select Segments")
        self.select_segments_chk.stateChanged.connect(self.toggle_select_segments)

        self.segment_dropdown = QComboBox()
        self.segment_dropdown.setEnabled(False)

        self.delete_segment_btn = QPushButton("Delete Segment")
        self.delete_segment_btn.setEnabled(False)
        self.delete_segment_btn.clicked.connect(self.delete_segment)

        self.select_segments_layout.addWidget(self.select_segments_chk)
        self.select_segments_layout.addWidget(self.segment_dropdown)
        self.select_segments_layout.addWidget(self.delete_segment_btn)

        bottom_right_layout.addWidget(self.calculate_avg_group)
        bottom_right_layout.addWidget(self.threshold_group)
        bottom_right_layout.addWidget(self.select_segments_group)

        bottom_button_layout = QHBoxLayout()
        bottom_right_layout.addLayout(bottom_button_layout)

        self.save_btn = QPushButton("Save Selected Data")
        self.save_btn.clicked.connect(self.save_selected_data)
        bottom_button_layout.addWidget(self.save_btn)
    
    def toggle_low_pass_filter(self, state):
        self.filter_type_dropdown.setEnabled(state == Qt.CheckState.Checked.value)
        self.cutoff_frequency_spinbox.setEnabled(state == Qt.CheckState.Checked.value)

    def select_folder(self):
        options = QFileDialog.Option.ShowDirsOnly
        directory = QFileDialog.getExistingDirectory(self, "Select Folder", "", options=options)
        if directory:
            self.folder_path_label.setText(f"Selected folder: {directory}")
            self.populate_file_list(directory, self.include_subfolders_chk.isChecked())

    def apply_low_pass_filter(self, data, cutoff_frequency, type, sampling_rate):
        if type == 'Butterworth':
            nyquist_rate = sampling_rate / 2.0
            cutoff = cutoff_frequency / nyquist_rate
            sos = butter(N=8, Wn=cutoff, btype='low', analog=False, output='sos')
            zi = sosfilt_zi(sos) * data[0]
            filtered_data, _ = sosfilt(sos, data, zi=zi)
            return filtered_data
        elif type == 'Bessel':
            nyquist_rate = sampling_rate / 2.0
            cutoff = cutoff_frequency / nyquist_rate
            sos = bessel(N=8, Wn=cutoff, btype='low', analog=False, output='sos')
            zi = sosfilt_zi(sos) * data[0]
            filtered_data, _ = sosfilt(sos, data, zi=zi)
            return filtered_data
        elif type == 'Chebyshev I':
            nyquist_rate = sampling_rate / 2.0
            cutoff = cutoff_frequency / nyquist_rate
            sos = cheby1(N=8, rp=0.1, Wn=cutoff, btype='low', analog=False, output='sos')
            zi = sosfilt_zi(sos) * data[0]
            filtered_data, _ = sosfilt(sos, data, zi=zi)
            return filtered_data
        elif type == 'Chebyshev II':
            nyquist_rate = sampling_rate / 2.0
            cutoff = cutoff_frequency / nyquist_rate
            sos = cheby2(N=8, rs=40, Wn=cutoff, btype='low', analog=False, output='sos')
            zi = sosfilt_zi(sos) * data[0]
            filtered_data, _ = sosfilt(sos, data, zi=zi)
            return filtered_data
        elif type == 'Elliptic':
            nyquist_rate = sampling_rate / 2.0
            cutoff = cutoff_frequency / nyquist_rate
            sos = ellip(N=8, rp=0.1, rs=40, Wn=cutoff, btype='low', analog=False, output='sos')
            zi = sosfilt_zi(sos) * data[0]
            filtered_data, _ = sosfilt(sos, data, zi=zi)
            return filtered_data
        elif type == 'FIR':
            nyquist_rate = sampling_rate / 2.0
            cutoff = cutoff_frequency / nyquist_rate
            taps = firwin(101, cutoff)
            filtered_data = lfilter(taps, 1, data)
            return filtered_data
        elif type == 'IIR':
            nyquist_rate = sampling_rate / 2.0
            cutoff = cutoff_frequency / nyquist_rate
            b, a = butter(N=8, Wn=cutoff, btype='low', analog=False)
            filtered_data = lfilter(b, a, data)
            return filtered_data

    def populate_file_list(self, directory, include_subfolders):
        self.files_list_widget.clear()
        for root, dirs, files in os.walk(directory):
            for file in files:
                if file.endswith('.abf') or file.endswith('.hdf5')or file.endswith('.h5'):
                    rel_path = os.path.relpath(os.path.join(root, file), start=directory)
                    item = QListWidgetItem(rel_path)
                    item.setData(Qt.ItemDataRole.UserRole, os.path.join(root, file))
                    self.files_list_widget.addItem(item)
            if not include_subfolders:
                break

    def plot_selected_file(self):
        selected_items = self.files_list_widget.selectedItems()
        if selected_items:
            file_path = selected_items[0].data(Qt.ItemDataRole.UserRole)
            if file_path.endswith('.abf'):
                self.data, self.sampling_rate, self.time = load_abf_file(file_path)
            elif file_path.endswith('.hdf5') or file_path.endswith('.h5'):
                self.data, self.sampling_rate, self.time = load_hdf5_file(file_path)

            if self.low_pass_filter_chk.isChecked():
                filter_type = self.filter_type_dropdown.currentText()
                cutoff_frequency = self.cutoff_frequency_spinbox.value() * 1000  # Convert kHz to Hz
                
                if self.sampling_rate <= cutoff_frequency * 2:
                    QMessageBox.warning(self, "Error", "The selected cutoff frequency is too high for the sampling rate.")
                    return
                
                self.data = self.apply_low_pass_filter(self.data, cutoff_frequency, filter_type, self.sampling_rate)
            
            # Clear existing plot and data
            self.canvas.axes.clear()
            self.segments.clear()
            self.segment_dropdown.clear()
            
            # Get the selected nth element value
            nth_element = self.nth_element_spinbox.value()
            
            # Plot every nth element of the data
            self.canvas.axes.plot(self.time[::nth_element], self.data[::nth_element], linewidth=0.5)
            self.canvas.axes.set_xlabel('Time (s)')
            self.canvas.axes.set_ylabel('Current (nA)')
            self.canvas.axes.set_xlim([min(self.time),max(self.time)])
            self.canvas.figure.tight_layout()
            self.canvas.draw()
            
            # Reset the states of checkboxes and inputs
            self.calculate_avg_chk.setChecked(False)
            self.threshold_chk.setChecked(False)
            self.select_segments_chk.setChecked(False)

    def update_plot(self):
        if self.data is not None:
            nth_element = self.nth_element_spinbox.value()
            if not self.canvas.axes.lines:
                self.canvas.axes.plot(self.time[::nth_element], self.data[::nth_element], linewidth=0.5)
                self.canvas.axes.set_xlabel('Time (s)')
                self.canvas.axes.set_ylabel('Current (nA)')
                self.canvas.axes.set_xlim([min(self.time),max(self.time)])
                self.canvas.figure.tight_layout()

            show_rolling_avg = self.calculate_avg_chk.isChecked()
            if show_rolling_avg:
                avg_window_size_in_ms = float(self.avg_window_size_input.text())
                rolling_avg = calculate_rolling_stats(self.data, self.sampling_rate, avg_window_size_in_ms)
                if len(self.canvas.axes.lines) > 1:
                    self.canvas.axes.lines[1].set_data(self.time, rolling_avg)
                else:
                    self.canvas.axes.plot(self.time, rolling_avg, linewidth=2)
            else:
                if len(self.canvas.axes.lines) > 1:
                    self.canvas.axes.lines[1].remove()

            show_threshold = self.threshold_chk.isChecked()
            if show_threshold:
                threshold_percentage = float(self.threshold_input.text()) / 100
                if self.data is not None:
                    if self.threshold_value_input.value() == 0:
                        self.threshold_value_input.setValue(int(np.mean(self.data)*1000))
                    threshold_value = self.threshold_value_input.value()/1000
                    avg_window_size_in_ms = float(self.avg_window_size_input.text())
                    rolling_avg = calculate_rolling_stats(self.data, self.sampling_rate, avg_window_size_in_ms)
                    threshold_lower = threshold_value - (threshold_value*(1-threshold_percentage))
                    threshold_upper = threshold_value + (threshold_value*(1-threshold_percentage))
                    
                    self.selected_regions = np.where((rolling_avg >= threshold_lower) & (rolling_avg <= threshold_upper))[0]
                    
                    # Remove existing threshold lines and spans
                    threshold_lines = [line for line in self.canvas.axes.lines if line.get_color() == 'r']
                    for line in threshold_lines:
                        line.remove()
                    threshold_spans = [patch for patch in self.canvas.axes.patches if patch.get_facecolor() == (1.0, 0.0, 0.0, 0.3)]
                    for span in threshold_spans:
                        span.remove()

                    # Add new threshold lines and spans
                    for region_start, region_end in self.find_contiguous_regions(self.selected_regions):
                        start_time = self.time[region_start]
                        end_time = self.time[region_end]
                        self.canvas.axes.axvline(x=start_time, color='r', linestyle='-', linewidth=0.5)
                        self.canvas.axes.axvline(x=end_time, color='r', linestyle='-', linewidth=0.5)
                        self.canvas.axes.axvspan(start_time, end_time, alpha=0.3, color='red')
            else:
                # Remove all threshold lines and spans
                threshold_lines = [line for line in self.canvas.axes.lines if line.get_color() == 'r']
                for line in threshold_lines:
                    line.remove()
                threshold_spans = [patch for patch in self.canvas.axes.patches if patch.get_facecolor() == (1.0, 0.0, 0.0, 0.3)]
                for span in threshold_spans:
                    span.remove()

            show_segments = self.select_segments_chk.isChecked()
            if show_segments:
                while len(self.canvas.axes.lines) > 2:
                    self.canvas.axes.lines[2].remove()
                for start_time, end_time in self.segments:
                    self.canvas.axes.axvline(x=start_time, color='g', linestyle='-', linewidth=0.5)
                    self.canvas.axes.axvline(x=end_time, color='g', linestyle='-', linewidth=0.5)
                    self.canvas.axes.axvspan(start_time, end_time, alpha=0.3, color='green')
            else:
                segment_lines = [line for line in self.canvas.axes.lines if line.get_color() == 'g']
                for line in segment_lines:
                    line.remove()
                segment_patches = [patch for patch in self.canvas.axes.patches if patch.get_facecolor() == (0.0, 0.5019607843137255, 0.0, 0.3)]
                for patch in segment_patches:
                    patch.remove()

            self.canvas.draw()



    def find_contiguous_regions(self, data):
        contiguous_regions = []
        data = np.sort(data)  # Sort the input data
        start_index = data[0]
        prev_index = start_index
        for i in range(1, len(data)):
            if data[i] > prev_index + 1:
                contiguous_regions.append((start_index, prev_index))
                start_index = data[i]
            prev_index = data[i]
        contiguous_regions.append((start_index, data[-1]))
        return contiguous_regions

    def toggle_avg_window_size(self, state):
        self.avg_window_size_input.setEnabled(state == Qt.CheckState.Checked.value)
        self.update_plot_btn.setEnabled(state == Qt.CheckState.Checked.value)
        self.update_plot()

    def toggle_threshold_input(self, state):
        self.threshold_input.setEnabled(state == Qt.CheckState.Checked.value)
        self.threshold_value_input.setEnabled(state == Qt.CheckState.Checked.value)
        self.select_segments_chk.setChecked(False)
        self.update_plot()

    def toggle_select_segments(self, state):
        if state == Qt.CheckState.Checked.value:
            segment_widget = SegmentWidget()
            segment_widget.add_button.clicked.connect(lambda: self.add_segment(segment_widget))
            self.segment_widgets.append(segment_widget)
            self.select_segments_layout.insertWidget(self.select_segments_layout.count() - 2, segment_widget)
            segment_widget.show()
            self.segment_dropdown.setEnabled(True)
            self.delete_segment_btn.setEnabled(True)
        else:
            while self.segment_widgets:
                segment_widget = self.segment_widgets.pop()
                self.select_segments_layout.removeWidget(segment_widget)
                segment_widget.deleteLater()
            self.segments.clear()
            self.segment_dropdown.clear()
            self.segment_dropdown.setEnabled(False)
            self.delete_segment_btn.setEnabled(False)

        self.threshold_chk.setChecked(False)
        self.update_plot()

    def add_segment(self, segment_widget):
        start_time = float(segment_widget.start_input.text())
        end_time = float(segment_widget.end_input.text())
        self.segments.append((start_time, end_time))
        segment_label = f"Segment: {start_time} - {end_time}"
        self.segment_dropdown.addItem(segment_label)
        index = self.segment_dropdown.count() - 1
        self.segment_dropdown.setItemData(index, segment_widget)  # Associate segment_widget with the dropdown item
        self.update_plot()

    def delete_segment(self):
        if self.segment_dropdown.count() > 0:
            index = self.segment_dropdown.currentIndex()
            self.segments.pop(index)
            segment_label = self.segment_dropdown.itemText(index)
            self.segment_dropdown.removeItem(index)
            for i in range(self.select_segments_layout.count()):
                widget = self.select_segments_layout.itemAt(i).widget()
                if isinstance(widget, QLabel) and widget.text() == segment_label:
                    self.select_segments_layout.removeWidget(widget)
                    widget.deleteLater()
                    break
            
            # Remove the segment lines and spans from the plot
            segment_lines = [line for line in self.canvas.axes.lines if line.get_color() == 'g']
            for line in segment_lines:
                line.remove()
            segment_patches = [patch for patch in self.canvas.axes.patches if patch.get_facecolor() == (0.0, 0.5019607843137255, 0.0, 0.3)]
            for patch in segment_patches:
                patch.remove()
            
            # Redraw the remaining segments
            for start_time, end_time in self.segments:
                self.canvas.axes.axvline(x=start_time, color='g', linestyle='-', linewidth=0.5)
                self.canvas.axes.axvline(x=end_time, color='g', linestyle='-', linewidth=0.5)
                self.canvas.axes.axvspan(start_time, end_time, alpha=0.3, color='green')
            
            self.canvas.draw()

    def save_selected_data(self):
        if self.data is not None:
            if self.threshold_chk.isChecked():
                selected_data = []
                for region_start, region_end in self.find_contiguous_regions(self.selected_regions):
                    start_index = np.searchsorted(self.time, self.time[region_start])
                    end_index = np.searchsorted(self.time, self.time[region_end])
                    selected_data.append(self.data[start_index:end_index+1])
                selected_data = np.concatenate(selected_data)
            elif self.select_segments_chk.isChecked():
                selected_data = []
                for start_time, end_time in self.segments:
                    start_index = np.searchsorted(self.time, start_time)
                    end_index = np.searchsorted(self.time, end_time)
                    selected_data.append(self.data[start_index:end_index+1])
                selected_data = np.concatenate(selected_data)
            else:
                QMessageBox.warning(self, "Warning", "No data selected.")
                return

            file_dialog = QFileDialog()
            file_dialog.setDefaultSuffix("h5")
            file_path, _ = file_dialog.getSaveFileName(self, "Save Selected Data", "", "HDF5 Files (*.h5)")

            if file_path:
                with h5py.File(file_path, 'w') as f:
                    f.create_dataset('selected_data', data=selected_data)
                    f.attrs['sampling_rate'] = self.sampling_rate
                QMessageBox.information(self, "Success", "Data saved successfully.")


if __name__ == '__main__':
    app = QApplication(sys.argv)
    app.setStyle(QStyleFactory.create('Fusion'))  # Use Fusion or other available styles
    # Customize the palette for a darker, more modern look
    palette = QPalette()
    palette.setColor(QPalette.ColorRole.Window, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.WindowText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Base, QColor(25, 25, 25))
    palette.setColor(QPalette.ColorRole.AlternateBase, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ToolTipBase, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.ToolTipText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Text, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ButtonText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.BrightText, Qt.GlobalColor.red)
    palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.HighlightedText, Qt.GlobalColor.black)
    app.setPalette(palette)
    window = MainWindow()
    window.showMaximized()
    sys.exit(app.exec())